/************************************************************************
 *   IRC - Internet Relay Chat, server/s_conf.c
 *
 *   Copyright (C) 2000-2003 TR-IRCD Development
 *
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: s_ip.c,v 1.1.1.1 2003/04/29 15:52:27 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "confitem.h"
#include "s_conf.h"
#include "h.h"

IP_ENTRY *free_ip_entries;
IP_ENTRY *ip_hash_table[IP_HASH_SIZE];

/*
 * usually, with hash tables, you use a prime number... but in this
 * case I am dealing with ip addresses, not ascii strings.
 */

void init_ip_hash(void);
static int hash_ip(struct irc_inaddr *ip_in);

/*
 * init_ip_hash()
 * 
 * input                - NONE output           - NONE side effects     - clear
 * the ip hash table
 * 
 * stole the link list pre-allocator from list.c
 */

void clear_ip_hash_table()
{
void *block_IP_ENTRIES;		/* block of IP_ENTRY's */
IP_ENTRY *new_IP_ENTRY;		/* new IP_ENTRY being made */
IP_ENTRY *last_IP_ENTRY;	/* last IP_ENTRY in chain */
int size, n_left_to_allocate = MAXCONNECTIONS;

    /*
     * ok. if the sizeof the struct isn't aligned with that of the
     * smallest guaranteed valid pointer (void *), then align it ya. you
     * could just turn 'size' into a #define. do it. :-)
     * 
     * -Dianora
     */

    size = sizeof(IP_ENTRY) + (sizeof(IP_ENTRY) & (sizeof(void *) - 1));

    block_IP_ENTRIES = (void *) MyMalloc((size * n_left_to_allocate));

    free_ip_entries = (IP_ENTRY *) block_IP_ENTRIES;
    last_IP_ENTRY = free_ip_entries;

    /*
     * *shudder* pointer arithmetic 
     */
    while (--n_left_to_allocate) {
	block_IP_ENTRIES = (void *) ((unsigned long) block_IP_ENTRIES + (unsigned long) size);
	new_IP_ENTRY = (IP_ENTRY *) block_IP_ENTRIES;
	last_IP_ENTRY->next = new_IP_ENTRY;
	new_IP_ENTRY->next = (IP_ENTRY *) NULL;
	last_IP_ENTRY = new_IP_ENTRY;
    }
    memset((void *) ip_hash_table, 0, sizeof(ip_hash_table));
}

/*
 * find_or_add_ip()
 * 
 * inputs               - unsigned long IP address value output         -
 * pointer to an IP_ENTRY element side effects  -
 * 
 * If the ip # was not found, a new IP_ENTRY is created, and the ip count
 * set to 0.
 */

IP_ENTRY *find_or_add_ip(struct irc_inaddr *ip_in)
{
    int hash_index;
    IP_ENTRY *ptr, *newptr;

    newptr = (IP_ENTRY *) NULL;
    for (ptr = ip_hash_table[hash_index = hash_ip(ip_in)]; ptr; ptr = ptr->next) {
	if (!memcmp(&ptr->ip, ip_in, sizeof(struct irc_inaddr))) {
	    return (ptr);
	}
    }

    if ((ptr = ip_hash_table[hash_index]) != (IP_ENTRY *) NULL) {
	if (free_ip_entries == (IP_ENTRY *) NULL)
	    outofmemory("Rehash IP hash");

	newptr = ip_hash_table[hash_index] = free_ip_entries;
	free_ip_entries = newptr->next;
	memcpy(&newptr->ip, ip_in, sizeof(*ip_in));
	newptr->count = 0;
	newptr->next = ptr;
	return (newptr);
    }
    if (free_ip_entries == (IP_ENTRY *) NULL)
	outofmemory("Free IP hash");

    ptr = ip_hash_table[hash_index] = free_ip_entries;
    free_ip_entries = ptr->next;
    memcpy(&ptr->ip, ip_in, sizeof(*ip_in));
    ptr->count = 0;
    ptr->next = (IP_ENTRY *) NULL;
    return (ptr);
}

/*
 * remove_one_ip
 * 
 * inputs               - unsigned long IP address value output         - NONE
 * side effects - ip address listed, is looked up in ip hash table and
 * number of ip#'s for that ip decremented. if ip # count reaches 0,
 * the IP_ENTRY is returned to the free_ip_enties link list.
 */

void remove_one_ip(struct irc_inaddr *ip_in)
{
    IP_ENTRY *ptr, **lptr;
    int hash_index = hash_ip(ip_in);
    for (lptr = ip_hash_table + hash_index, ptr = *lptr; ptr; lptr = &ptr->next, ptr = *lptr) {
	if (memcmp(&IN_ADDR(ptr->ip), &PIN_ADDR(ip_in), sizeof(struct irc_inaddr)))
	    continue;
	if (ptr->count != 0)
	    ptr->count--;
	if (ptr->count != 0)
	    continue;
	*lptr = ptr->next;
	ptr->next = free_ip_entries;
	free_ip_entries = ptr;
	return;
    }
}

/*
 * hash_ip()
 * 
 * input        - pointer to an irc_inaddr
 * output       - integer value used as index into hash table
 * side effects - hopefully, none
 */

#ifndef IPV6
static int hash_ip(struct irc_inaddr *addr)
{
    int hash;
    u_int32_t ip;

    ip = ntohl(PIN_ADDR(addr));
    hash = ((ip >> 12) + ip) & (IP_HASH_SIZE - 1);
    return (hash);
}
#else /* IPV6 */
static int hash_ip(struct irc_inaddr *addr)
{
    int hash;
    u_int32_t *ip = (u_int32_t *) & PIN_ADDR(addr);

    if (IN6_IS_ADDR_V4MAPPED((struct in6_addr *) ip)) {
	hash = ((ip[3] >> 12) + ip[3]) & (IP_HASH_SIZE - 1);
	return (hash);
    }

    hash = ip[0] ^ ip[3];
    hash ^= hash >> 16;
    hash ^= hash >> 8;
    hash = hash & (IP_HASH_SIZE - 1);
    return (hash);
}
#endif /* IPV6 */

/*
 * count_ip_hash
 * 
 * inputs               - pointer to counter of number of ips hashed - pointer
 * to memory used for ip hash output            - returned via pointers
 * input side effects   - NONE
 * 
 * number of hashed ip #'s is counted up, plus the amount of memory used
 * in the hash.
 * Added so s_debug could check memory usage in here -Dianora 
 */

void count_ip_hash(int *number_ips_stored, u_long *mem_ips_stored)
{
    IP_ENTRY *ip_hash_ptr;
    int i;

    *number_ips_stored = 0;
    *mem_ips_stored = 0;

    for (i = 0; i < IP_HASH_SIZE; i++) {
	ip_hash_ptr = ip_hash_table[i];
	while (ip_hash_ptr) {
	    *number_ips_stored = *number_ips_stored + 1;
	    *mem_ips_stored = *mem_ips_stored + sizeof(IP_ENTRY);

	    ip_hash_ptr = ip_hash_ptr->next;
	}
    }
}

/*
 * iphash_stats()
 *
 * input        - 
 * output       -
 * side effects -
 */
void
iphash_stats(struct Client *client_p, struct Client *source_p, int parc, char *parv[], FBFILE * out)
{
    IP_ENTRY *ip_hash_ptr;
    int i;
    int collision_count;
    char result_buf[256];

    if (out == NULL)
	send_me_debug(source_p, ":*** hash stats for iphash");
    else {
	ircsprintf(result_buf, "*** hash stats for iphash\n");
	fbputs(result_buf, out);
    }

    for (i = 0; i < IP_HASH_SIZE; i++) {
	ip_hash_ptr = ip_hash_table[i];

	collision_count = 0;
	while (ip_hash_ptr) {
	    collision_count++;
	    ip_hash_ptr = ip_hash_ptr->next;
	}
	if (collision_count) {
	    if (out == NULL) {
		send_me_debug(source_p, ":Entry %d (0x%X) Collisions %d", i, i, collision_count);
	    } else {
		ircsprintf(result_buf, "Entry %d (0x%X) Collisions %d\n", i, i, collision_count);
		fbputs(result_buf, out);
	    }
	}
    }
}
